from os import PathLike
import os.path
from typing import List, Union, BinaryIO
import tarfile
import zipfile
from fnmatch import fnmatch, fnmatchcase


ARCHIVE_EXTENSIONS = ['ZIP', 'zip', 'TAR', 'tar', 'TAR.BZ2', 'tar.bz2', 'TAR.GZ', 'tar.gz']

def filter_filenames(filenames: List[PathLike], glob: str, case=False) -> List[PathLike]:
    cmp = fnmatchcase if case else fnmatch
    return [
        filename
        for filename in filenames
        if cmp(filename, glob)
    ]


def is_tarfile(archive_file: Union[PathLike, BinaryIO]) -> bool:
    """ Helper to detect whether a path or a file object is
        referencing a valid TAR file.
    """
    try:
        return tarfile.is_tarfile(archive_file)
    except TypeError:
        pass

    try:
        tarfile.open(fileobj=archive_file)
        return True
    except (TypeError, tarfile.ReadError):
        return False

def open_tarfile(archive_file: Union[PathLike, BinaryIO]) -> tarfile.TarFile:
    """ Open a TAR file from either a path or a file object.
    """
    if isinstance(archive_file, BinaryIO):
        return tarfile.open(fileobj=archive_file)
    return tarfile.open(archive_file)


def unpack_files(archive_path: Union[PathLike, BinaryIO] , target_dir: PathLike, glob=None, filenames=None, recursive=False) -> List[PathLike]:
    """ Unpacks the contents of the specified ZIP or TAR archive to the
        given target directory. Optionally, only a given list of filenames
        will be extracted.
        When a glob is passed, all filenames (either given or from the archive)
        will be filtered and only the matching files will be extracted.
    """
    iszip = False
    istar = False
    # open the archive and extract a list of filenames
    if is_tarfile(archive_path):
        archive = open_tarfile(archive_path)
        all_filenames = archive.getnames()
        filenames = filenames or all_filenames
        istar = True
    elif zipfile.is_zipfile(archive_path):
        archive = zipfile.ZipFile(archive_path)
        all_filenames = archive.namelist()
        filenames = filenames or all_filenames
        iszip = True
    else:
        raise Exception('Cannot open archive %s' % archive_path)

    # filter the filenames when a glob is passed
    if glob:
        filenames = filter_filenames(filenames, glob)

    extracted_filenames = []

    # extract the files to the target directory
    if istar:
        members = [
            member
            for member in archive.getmembers()
            if member.name in filenames
        ]
        archive.extractall(target_dir, members)
        extracted_filenames.extend([
            os.path.join(target_dir, member.name)
            for member in members
        ])

    elif iszip:
        archive.extractall(target_dir, filenames)
        extracted_filenames.extend([
            os.path.join(target_dir, filename)
            for filename in filenames
        ])

    # go into the sub-archives to extract files
    if recursive:
        for extension in ARCHIVE_EXTENSIONS:
            sub_archives = filter_filenames(all_filenames, '*.%s' % extension)
            for sub_archive in sub_archives:
                if istar:
                    sub_archive_file = archive.extractfile(
                        archive.getmember(sub_archive)
                    )
                if iszip:
                    sub_archive_file = archive.open(sub_archive)

                sub_filenames = unpack_files(
                    sub_archive_file,
                    os.path.join(target_dir, sub_archive),
                    glob,
                    filenames,
                    recursive
                )
                extracted_filenames.extend(sub_filenames)

    # return a list of files extracted
    return extracted_filenames